# encoding: utf-8

from collections import OrderedDict
from contextlib import contextmanager

import six
from django.utils.tree import Node

from .compat import (
    ADD_Q_METHOD_NAME, ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP, BUILD_FILTER_METHOD_NAME, NAMES_TO_PATH_METHOD_NAME,
    NEED_HAVING_METHOD_NAME, QUERY_CHAIN_METHOD_NAME, ValuesQuerySet, contains_aggregate,
    convert_build_filter_to_add_q_kwargs, nullcontext,
)
from .exceptions import QueryablePropertyError
from .utils.internal import InjectableMixin, NodeChecker, QueryPath, resolve_queryable_property

QUERYING_PROPERTIES_MARKER = '__querying_properties__'


class AggregatePropertyChecker(NodeChecker):
    """
    A specialized node checker that checks whether or not a node contains a
    reference to an aggregate property for the purposes of determining whether
    or not a HAVING clause is required.
    """

    def __init__(self):
        super(AggregatePropertyChecker, self).__init__(self.is_aggregate_property)

    def is_aggregate_property(self, item, model, ignored_refs=frozenset()):
        """
        Check if the given node item or its subnodes contain a reference to an
        aggregate property.

        :param (str, object) item: The node item consisting of path and value.
        :param model: The model class the corresponding query is performed for.
        :param ignored_refs: Queryable property references that should not be
                             checked.
        :type ignored_refs: frozenset[queryable_properties.utils.internal.QueryablePropertyReference]
        :return:
        """
        property_ref, lookups = resolve_queryable_property(model, QueryPath(item[0]))
        if not property_ref or property_ref in ignored_refs:
            return False
        if property_ref.property.filter_requires_annotation:
            if contains_aggregate(property_ref.get_annotation()):
                return True
            ignored_refs = ignored_refs.union((property_ref,))
        # Also check the Q object returned by the property's get_filter method
        # as it may contain references to other properties that may add
        # aggregation-based annotations.
        return self.check_leaves(property_ref.get_filter(lookups, item[1]), model=model, ignored_refs=ignored_refs)


aggregate_property_checker = AggregatePropertyChecker()


class QueryablePropertiesCompilerMixin(InjectableMixin):
    """
    A mixin for :class:`django.db.models.sql.compiler.SQLCompiler` objects that
    extends the original Django objects to inject the
    ``QUERYING_PROPERTIES_MARKER``.
    """

    def setup_query(self, *args, **kwargs):
        super(QueryablePropertiesCompilerMixin, self).setup_query(*args, **kwargs)
        # Add the marker to the column map while ensuring that it's the first
        # entry.
        annotation_col_map = OrderedDict()
        annotation_col_map[QUERYING_PROPERTIES_MARKER] = -1
        annotation_col_map.update(self.annotation_col_map)
        self.annotation_col_map = annotation_col_map

    def results_iter(self, *args, **kwargs):
        for row in super(QueryablePropertiesCompilerMixin, self).results_iter(*args, **kwargs):
            # Add the fixed value for the fake querying properties marker
            # annotation to each row. In recent versions, the value can simply
            # be appended since -1 can be specified as the index in the
            # annotation_col_map. In old versions, the value must be injected
            # as the first annotation value.
            addition = row.__class__((True,))
            if not ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:
                row += addition
            else:  # pragma: no cover
                index = len(row) - len(self.query.aggregate_select) - len(self.query.related_select_cols)
                row = row[:index] + addition + row[index:]
            yield row


class QueryablePropertiesQueryMixin(InjectableMixin):
    """
    A mixin for :class:`django.db.models.sql.Query` and
    :class:`django.db.models.sql.Raw Query` objects that extends the original
    Django objects to deal with queryable properties, e.g. managing used
    properties or automatically adding required properties as annotations.
    """

    def __getattr__(self, name):  # pragma: no cover
        # Redirect some attribute accesses for older Django versions (where
        # annotations were tied to aggregations, hence "aggregation" in the
        # names instead of "annotation").
        if name in ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:
            return getattr(self, ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP[name])
        raise AttributeError()

    def __iter__(self):  # Raw queries
        # See QueryablePropertiesCompilerMixin.results_iter, but for raw
        # queries. The marker can simply be added as the first value as fields
        # are not strictly grouped like in regular queries.
        for row in super(QueryablePropertiesQueryMixin, self).__iter__():
            if self._use_querying_properties_marker:
                row = row.__class__((True,)) + row
            yield row

    def init_injected_attrs(self):
        # Stores references to queryable properties used as annotations in this
        # query.
        self._queryable_property_annotations = set()
        # A stack for queryable properties who are currently being annotated.
        # Required to correctly resolve dependencies and perform annotations.
        self._queryable_property_stack = []
        # Determines whether to inject the QUERYING_PROPERTIES_MARKER.
        self._use_querying_properties_marker = False

    @contextmanager
    def _add_queryable_property_annotation(self, property_ref, full_group_by, select=False):
        """
        A context manager that adds a queryable property annotation to this
        query and performs management tasks around the annotation (stores the
        information if the queryable property annotation should be selected
        and populates the queryable property stack correctly). The context
        manager yields the actual resolved and applied annotation while the
        stack is still populated.

        :param property_ref: A reference containing the queryable property
                             to annotate.
        :type property_ref: queryable_properties.utils.internal.QueryablePropertyReference
        :param bool full_group_by: Signals whether to use all fields of the
                                   query for the GROUP BY clause when dealing
                                   with an aggregate-based annotation or not.
        :param bool select: Signals whether the annotation should be selected
                            or not.
        """
        if property_ref in self._queryable_property_stack:
            raise QueryablePropertyError('Queryable property "{}" has a circular dependency and requires itself.'
                                         .format(property_ref.property))

        annotation_name = six.text_type(property_ref.full_path)
        annotation_mask = set(self.annotations if self.annotation_select_mask is None else self.annotation_select_mask)
        self._queryable_property_stack.append(property_ref)
        try:
            if property_ref not in self._queryable_property_annotations:
                self.add_annotation(property_ref.get_annotation(), alias=annotation_name)
                if not select:
                    self.set_annotation_mask(annotation_mask)
                self._queryable_property_annotations.add(property_ref)
            elif select and self.annotation_select_mask is not None:
                self.set_annotation_mask(annotation_mask.union((annotation_name,)))
            annotation = self.annotations[annotation_name]
            yield annotation
        finally:
            self._queryable_property_stack.pop()

        # Perform the required GROUP BY setup if the annotation contained
        # aggregates, which is normally done by QuerySet.annotate.
        if contains_aggregate(annotation):
            if full_group_by and not ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:
                # In recent Django versions, a full GROUP BY can be achieved by
                # simply setting group_by to True.
                self.group_by = True
            else:
                if full_group_by and self.group_by is None:  # pragma: no cover
                    # In old versions, the fields must be added to the selected
                    # fields manually and set_group_by must be called after.
                    opts = self.model._meta
                    self.add_fields([f.attname for f in getattr(opts, 'concrete_fields', opts.fields)], False)
                self.set_group_by()

    def _auto_annotate(self, query_path, full_group_by=None):
        """
        Try to resolve the given path into a queryable property and annotate
        the property as a non-selected property (if the property wasn't added
        as an annotation already). Do nothing if the path does not match a
        queryable property.

        :param QueryPath query_path: The query path to resolve.
        :param bool | None full_group_by: Optional override to indicate whether
                                          or not all fields must be contained
                                          in a GROUP BY clause for aggregate
                                          annotations. If not set, it will be
                                          determined from the state of this
                                          query.
        :return: The resolved annotation or None if the path couldn't be
                 resolved.
        """
        property_ref = resolve_queryable_property(self.model, query_path)[0]
        if not property_ref:
            return None
        if full_group_by is None:
            full_group_by = bool(ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP) and not self.select
        with self._add_queryable_property_annotation(property_ref, full_group_by) as annotation:
            return annotation

    def _postprocess_clone(self, clone):
        """
        Postprocess a query that was the result of cloning this query. This
        ensures that the cloned query also uses this mixin and that the
        queryable property attributes are initialized correctly.

        :param django.db.models.sql.Query clone: The cloned query.
        :return: The postprocessed cloned query.
        :rtype: django.db.models.sql.Query
        """
        QueryablePropertiesQueryMixin.inject_into_object(clone)
        clone.init_injected_attrs()
        clone._queryable_property_annotations.update(self._queryable_property_annotations)
        return clone

    def add_aggregate(self, aggregate, model=None, alias=None, is_summary=False):  # pragma: no cover
        # This method is called in older versions to add an aggregate, which
        # may be based on a queryable property annotation, which in turn must
        # be auto-annotated here.
        query_path = QueryPath(aggregate.lookup)
        if self._queryable_property_stack:
            query_path = self._queryable_property_stack[-1].relation_path + query_path
        property_annotation = self._auto_annotate(query_path)
        if property_annotation:
            # If it is based on a queryable property annotation, annotating the
            # current aggregate cannot be delegated to Django as it couldn't
            # deal with annotations containing the lookup separator.
            aggregate.add_to_query(self, alias, six.text_type(query_path), property_annotation, is_summary)
        else:
            # The overridden method also allows to set a default value for the
            # model parameter, which will be missing if add_annotation calls are
            # redirected to add_aggregate for older Django versions.
            model = model or self.model
            super(QueryablePropertiesQueryMixin, self).add_aggregate(aggregate, model, alias, is_summary)
        if self.annotation_select_mask is not None:
            self.set_annotation_mask(self.annotation_select_mask.union((alias,)))

    def add_filter(self, *args, **kwargs):  # pragma: no cover
        # The build_filter method was called add_filter in very old Django
        # versions. Since recent versions still have an add_filter method (for
        # different purposes), the queryable properties customizations should
        # only occur in old versions.
        if BUILD_FILTER_METHOD_NAME == 'add_filter':
            # Simply use the build_filter implementation that does all the
            # heavy lifting and is aware of the different methods in different
            # versions and therefore calls the correct super methods if
            # necessary.
            return self.build_filter(*args, **kwargs)
        return super(QueryablePropertiesQueryMixin, self).add_filter(*args, **kwargs)

    def add_ordering(self, *ordering, **kwargs):
        for field_name in ordering:
            # Ordering by a queryable property via simple string values
            # requires auto-annotating here, while a queryable property used
            # in a complex ordering expression is resolved through other
            # overridden methods.
            if isinstance(field_name, six.string_types) and field_name != '?':
                if field_name.startswith('-'):
                    field_name = field_name[1:]
                self._auto_annotate(QueryPath(field_name))
        return super(QueryablePropertiesQueryMixin, self).add_ordering(*ordering, **kwargs)

    @property
    def aggregate_select(self):  # pragma: no cover
        select = original = super(QueryablePropertiesQueryMixin, self).aggregate_select
        if self._use_querying_properties_marker:
            # Since old Django versions don't offer the annotation_col_map on
            # compilers, but read the annotations directly from the query, the
            # querying properties marker has to be injected here. The value for
            # the annotation will be provided via the compiler mixin.
            select = OrderedDict()
            select[QUERYING_PROPERTIES_MARKER] = None
            select.update(original)
        return select

    def build_filter(self, filter_expr, *args, **kwargs):
        # Check if the given filter expression is meant to use a queryable
        # property. Therefore, the possibility of filter_expr not being of the
        # correct type must be taken into account (a case Django would cover
        # already, but the check for queryable properties MUST run first).
        try:
            arg, value = filter_expr
        except (TypeError, ValueError):
            # Invalid value - just treat it as "no queryable property found"
            # and delegate it to Django.
            property_ref = None
        else:
            property_ref, lookups = resolve_queryable_property(self.model, QueryPath(arg))

        # If no queryable property could be determined for the filter
        # expression (either because a regular/non-existent field is referenced
        # or because the expression was a special or invalid value), call
        # Django's default implementation, which may in turn raise an
        # exception. Act the same way if the current top of the stack is used
        # to avoid infinite recursions.
        if not property_ref or (self._queryable_property_stack and self._queryable_property_stack[-1] == property_ref):
            # The base method has different names in different Django versions
            # (see comment on the constant definition).
            base_method = getattr(super(QueryablePropertiesQueryMixin, self), BUILD_FILTER_METHOD_NAME)
            return base_method(filter_expr, *args, **kwargs)

        q_obj = property_ref.get_filter(lookups, value)
        # Before applying the filter implemented by the property, check if
        # the property signals the need of its own annotation to function.
        # If so, add the annotation first to avoid endless recursion, since
        # resolved filter will likely contain the same property name again.
        context = nullcontext()
        if property_ref.property.filter_requires_annotation:
            full_group_by = bool(ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP) and not self.select
            context = self._add_queryable_property_annotation(property_ref, full_group_by)

        with context:
            # Luckily, build_filter and _add_q use the same return value
            # structure, so an _add_q call can be used to actually create the
            # return value for the current call. The (_)add_q method has
            # different names in different Django versions (see comment on the
            # constant definition).
            method = getattr(self, ADD_Q_METHOD_NAME)
            return method(q_obj, **convert_build_filter_to_add_q_kwargs(**kwargs))

    def get_aggregation(self, *args, **kwargs):
        # If the query is to be used as a pure aggregate query (which might use
        # a subquery), all queryable property annotations must be added to the
        # select mask to avoid potentially empty SELECT clauses.
        if self.annotation_select_mask is not None and self._queryable_property_annotations:
            annotation_names = (six.text_type(property_ref.full_path) for property_ref
                                in self._queryable_property_annotations)
            self.set_annotation_mask(set(self.annotation_select_mask).union(annotation_names))
        return super(QueryablePropertiesQueryMixin, self).get_aggregation(*args, **kwargs)

    def get_columns(self):  # Raw queries
        # Like QueryablePropertiesCompilerMixin.setup_query, but for raw
        # queries. The marker can simply be added as the first value as fields
        # are not strictly grouped like in regular queries.
        columns = super(QueryablePropertiesQueryMixin, self).get_columns()
        if self._use_querying_properties_marker:
            columns.insert(0, QUERYING_PROPERTIES_MARKER)
        return columns

    def get_compiler(self, *args, **kwargs):
        use_marker = self._use_querying_properties_marker
        self._use_querying_properties_marker = False
        compiler = super(QueryablePropertiesQueryMixin, self).get_compiler(*args, **kwargs)
        if use_marker:
            QueryablePropertiesCompilerMixin.inject_into_object(compiler)
        return compiler

    def names_to_path(self, names, *args, **kwargs):
        # This is a central method for resolving field names. To also allow the
        # use of queryable properties across relations, the relation path on
        # top of the stack must be prepended to trick Django into resolving
        # correctly.
        if self._queryable_property_stack:
            names = self._queryable_property_stack[-1].relation_path + names
        base_method = getattr(super(QueryablePropertiesQueryMixin, self), NAMES_TO_PATH_METHOD_NAME)
        return base_method(names, *args, **kwargs)

    def need_force_having(self, q_object):  # pragma: no cover
        # Same as need_having, but for even older versions. Simply delegate to
        # need_having, which is aware of the different methods in different
        # versions and therefore calls the correct super method if necessary.
        return self.need_having(q_object)

    def need_having(self, obj):  # pragma: no cover
        # This method is used by older Django versions to figure out if the
        # filter represented by a Q object must be put in the HAVING clause of
        # the query. Since a queryable property might add an aggregate-based
        # annotation during the actual filter application, this method must
        # return True if a filter condition contains such a property.
        node = obj if isinstance(obj, Node) else Node([obj])
        if aggregate_property_checker.check_leaves(node, model=self.model):
            return True
        # The base method has different names in different Django versions (see
        # comment on the constant definition).
        base_method = getattr(super(QueryablePropertiesQueryMixin, self), NEED_HAVING_METHOD_NAME)
        return base_method(obj)

    def resolve_ref(self, name, allow_joins=True, reuse=None, summarize=False, *args, **kwargs):
        # This method is used to resolve field names in complex expressions. If
        # a queryable property is used in such an expression, it needs to be
        # auto-annotated (while taking the stack into account) and returned.
        query_path = QueryPath(name)
        if self._queryable_property_stack:
            query_path = self._queryable_property_stack[-1].relation_path + query_path
        property_annotation = self._auto_annotate(query_path, full_group_by=ValuesQuerySet is not None)
        if property_annotation:
            if summarize:
                # Outer queries for aggregations need refs to annotations of
                # the inner queries.
                from django.db.models.expressions import Ref
                return Ref(name, property_annotation)
            return property_annotation
        return super(QueryablePropertiesQueryMixin, self).resolve_ref(name, allow_joins, reuse, summarize,
                                                                      *args, **kwargs)

    def setup_joins(self, names, *args, **kwargs):
        # This method contained the logic of names_to_path in very old Django
        # versions. Simply delegate to the overridden names_to_path in this
        # case, which is aware of the different methods in different versions
        # and therefore calls the correct super method.
        if NAMES_TO_PATH_METHOD_NAME == 'setup_joins':  # pragma: no cover
            return self.names_to_path(names, *args, **kwargs)
        return super(QueryablePropertiesQueryMixin, self).setup_joins(names, *args, **kwargs)

    def clone(self, *args, **kwargs):
        obj = super(QueryablePropertiesQueryMixin, self).clone(*args, **kwargs)
        if QUERY_CHAIN_METHOD_NAME == 'clone':  # pragma: no cover
            obj = self._postprocess_clone(obj)
        return obj

    def chain(self, *args, **kwargs):
        obj = super(QueryablePropertiesQueryMixin, self).chain(*args, **kwargs)
        if QUERY_CHAIN_METHOD_NAME == 'chain':
            obj = self._postprocess_clone(obj)
        return obj



# encoding: utf-8

import pytest
from django import VERSION as DJANGO_VERSION
from mock import Mock, patch
from six.moves import cPickle

from queryable_properties.compat import ModelIterable, ValuesQuerySet
from queryable_properties.managers import (
    LegacyIterable, LegacyOrderingMixin, LegacyOrderingModelIterable, LegacyValuesIterable, LegacyValuesListIterable,
    QueryablePropertiesIterableMixin, QueryablePropertiesManager, QueryablePropertiesManagerMixin,
    QueryablePropertiesQuerySet, QueryablePropertiesQuerySetMixin,
)
from queryable_properties.query import QUERYING_PROPERTIES_MARKER
from queryable_properties.utils import get_queryable_property
from queryable_properties.utils.internal import QueryablePropertyReference, QueryPath
from .app_management.models import (
    ApplicationTag, ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
    VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties,
)

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class DummyIterable(QueryablePropertiesIterableMixin, ModelIterable or LegacyIterable):
    pass


class DummyOrderingIterable(LegacyOrderingMixin, LegacyIterable):
    pass


@pytest.fixture
def refs():
    model = ApplicationWithClassBasedProperties
    return {
        prop_name: QueryablePropertyReference(get_queryable_property(model, prop_name), model, QueryPath())
        for prop_name in ('major_sum', 'version_count')
    }


class TestQueryablePropertiesQuerySetMixin(object):

    def assert_queryset_picklable(self, queryset, selected_descriptors=()):
        expected_results = list(queryset)
        serialized_queryset = cPickle.dumps(queryset)
        deserialized_queryset = cPickle.loads(serialized_queryset)
        assert list(deserialized_queryset) == expected_results
        for descriptor in selected_descriptors:
            assert all(descriptor.has_cached_value(obj) for obj in deserialized_queryset)

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_pickle_model_instance_queryset(self, model):
        queryset = model.objects.filter(version_count=4).order_by('name').select_properties('version_count')
        self.assert_queryset_picklable(queryset, (model.version_count,))

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_pickle_values_queryset(self, model):
        queryset = model.objects.order_by('-pk').select_properties('version_count').values('name', 'version_count')
        self.assert_queryset_picklable(queryset)

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_pickle_values_list_queryset(self, model):
        queryset = model.objects.order_by('pk').select_properties('version_count').values_list('name', 'version_count')
        self.assert_queryset_picklable(queryset)

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_pickle_dates_queryset(self, model):
        queryset = model.objects.filter(application__version_count=3).dates('supported_from', 'year')
        self.assert_queryset_picklable(queryset)

    @pytest.mark.skipif(DJANGO_VERSION >= (1, 9), reason="_clone doesn't change the class in recent Django versions.")
    @pytest.mark.parametrize('change_class, setup, kwargs', [
        (False, False, {}),
        (False, True, {'dummy': None}),
        (True, False, {}),
        (True, True, {'test1': 'test', 'test2': 1337}),
    ])
    def test_clone_with_class_change(self, change_class, setup, kwargs):
        queryset = ApplicationWithClassBasedProperties.objects.all()
        assert isinstance(queryset, QueryablePropertiesQuerySetMixin)
        cls = ValuesQuerySet if change_class else None
        clone = queryset._clone(cls, setup, _fields=[], **kwargs)
        assert isinstance(clone, ValuesQuerySet) is change_class
        for name, value in kwargs.items():
            assert getattr(clone, name) == value

    def test_apply_to(self, tags):
        queryset_without_properties = ApplicationTag.objects.all()
        assert not isinstance(queryset_without_properties, QueryablePropertiesQuerySetMixin)

        queryset = QueryablePropertiesQuerySetMixin.apply_to(queryset_without_properties)
        assert queryset is not queryset_without_properties
        assert isinstance(queryset, QueryablePropertiesQuerySetMixin)
        assert list(queryset_without_properties) == list(queryset)
        assert set(queryset.filter(applications__version_count=4)) == set(tags)


class TestQueryablePropertiesQuerySet(object):

    def test_get_for_model(self, tags):
        queryset_without_properties = ApplicationTag._default_manager.all()
        assert not isinstance(queryset_without_properties, QueryablePropertiesQuerySetMixin)

        queryset = QueryablePropertiesQuerySet.get_for_model(ApplicationTag)
        assert isinstance(queryset, QueryablePropertiesQuerySetMixin)
        assert list(queryset_without_properties) == list(queryset)
        assert set(queryset.filter(applications__version_count=4)) == set(tags)


class TestQueryablePropertiesManagerMixin(object):

    def test_apply_to(self, tags):
        assert not isinstance(ApplicationTag.objects, QueryablePropertiesManagerMixin)

        manager = QueryablePropertiesManagerMixin.apply_to(ApplicationTag.objects)
        assert manager is not ApplicationTag.objects
        assert isinstance(manager, QueryablePropertiesManagerMixin)
        assert manager.model is ApplicationTag
        assert manager._db == ApplicationTag.objects._db
        assert getattr(manager, '_hints', None) == getattr(ApplicationTag.objects, '_hints', None)
        assert manager.name == '<{}_with_queryable_properties>'.format(
            getattr(ApplicationTag.objects, 'name', 'manager'))

        queryset = manager.all()
        assert isinstance(queryset, QueryablePropertiesQuerySetMixin)
        assert set(queryset.filter(applications__version_count=4)) == set(tags)


class TestQueryablePropertiesManager(object):

    @pytest.mark.parametrize('using, hints', [
        (None, None),
        (None, {}),
        ('default', None),
        ('default', {'test': 'hint'}),
    ])
    def test_get_for_model(self, tags, using, hints):
        assert not isinstance(ApplicationTag._default_manager, QueryablePropertiesManagerMixin)

        manager = QueryablePropertiesManager.get_for_model(ApplicationTag, using, hints)
        assert isinstance(manager, QueryablePropertiesManagerMixin)
        assert manager.model is ApplicationTag
        assert manager._db == using
        assert manager._hints == (hints or {})
        assert manager.name == '<manager_with_queryable_properties>'

        queryset = manager.all()
        assert isinstance(queryset, QueryablePropertiesQuerySetMixin)
        assert set(queryset.filter(applications__version_count=4)) == set(tags)


class TestLegacyIterable(object):

    def test_initializer(self):
        queryset = ApplicationWithClassBasedProperties.objects.all()
        iterable = LegacyIterable(queryset)
        assert iterable.queryset is queryset

    def test_iter(self):
        for queryset in (
            ApplicationWithClassBasedProperties.objects.order_by('pk'),
            ApplicationWithClassBasedProperties.objects.order_by('pk').values('pk', 'name'),
            ApplicationWithClassBasedProperties.objects.order_by('pk').values_list('pk', 'name'),
            VersionWithClassBasedProperties.objects.dates('supported_from', 'year'),
        ):
            iterable = LegacyIterable(queryset)
            assert list(iterable) == list(queryset)


class TestQueryablePropertiesIterableMixin(object):

    def test_initializer(self):
        queryset = ApplicationWithClassBasedProperties.objects.order_by('pk')
        iterable = DummyIterable(queryset)
        assert iterable.queryset is not queryset
        assert list(queryset) == list(iterable.queryset)

    def test_postprocess_queryable_properties(self):
        iterable = DummyIterable(ApplicationWithClassBasedProperties.objects.order_by('pk'))
        application = ApplicationWithClassBasedProperties()
        assert iterable._postprocess_queryable_properties(application) is application

    def test_iter(self):
        queryset = ApplicationWithClassBasedProperties.objects.order_by('pk')
        iterable = DummyIterable(queryset)
        mock_setup = Mock()
        mock_postprocess = Mock(side_effect=lambda obj: obj)
        with patch.multiple(iterable, _setup_queryable_properties=mock_setup,
                            _postprocess_queryable_properties=mock_postprocess):
            applications = list(iterable)
        assert applications == list(queryset)
        mock_setup.assert_called_once_with()
        assert mock_postprocess.call_count == len(applications)
        for application in applications:
            mock_postprocess.assert_any_call(application)


class TestLegacyOrderingMixin(object):

    @pytest.mark.parametrize('order_by, expected_indexes', [
        ((), {}),
        (('name', '-pk'), {}),
        (('version_count', '-pk'), {'version_count': [0]}),
        (('name', '-major_sum'), {'major_sum': [1]}),
        (('major_sum', '-version_count', 'name', '-major_sum'), {'major_sum': [0, 3], 'version_count': [1]}),
    ])
    def test_order_by_occurrences(self, order_by, expected_indexes):
        queryset = ApplicationWithClassBasedProperties.objects.order_by(*order_by)
        iterable = DummyOrderingIterable(queryset)
        assert len(iterable._order_by_occurrences) == len(expected_indexes)
        for ref, indexes in iterable._order_by_occurrences.items():
            assert expected_indexes[ref.property.name] == indexes

    @pytest.mark.parametrize('order_by, select, expected_result', [
        ((), (), set()),
        (('name', '-pk'), (), set()),
        (('version_count', '-pk'), (), {'version_count'}),
        (('name', '-major_sum'), ('major_sum',), set()),
        (('major_sum', '-version_count', 'name', '-major_sum'), (), {'major_sum', 'version_count'}),
        (('major_sum', '-version_count', 'name', '-major_sum'), ('version_count',), {'major_sum'}),
        (('major_sum', '-version_count', 'name', '-major_sum'), ('version_count', 'major_sum'), set()),
    ])
    def test_order_by_select(self, order_by, select, expected_result):
        queryset = ApplicationWithClassBasedProperties.objects.select_properties(*select).order_by(*order_by)
        iterable = DummyOrderingIterable(queryset)
        assert {ref.property.name for ref in iterable._order_by_select} == expected_result

    @pytest.mark.parametrize('order_by_select', [
        (),
        ('version_count',),
        ('major_sum', 'version_count'),
    ])
    def test_setup_queryable_properties(self, refs, order_by_select):
        queryset = ApplicationWithClassBasedProperties.objects.order_by('-major_sum', 'version_count')
        iterable = DummyOrderingIterable(queryset)
        iterable.__dict__['_order_by_select'] = {refs[prop_name] for prop_name in order_by_select}
        iterable._setup_queryable_properties()
        query = iterable.queryset.query
        for prop_name in ('major_sum', 'version_count'):
            if prop_name in order_by_select:
                assert query.annotation_select[prop_name] == query.annotations[prop_name]
            else:
                assert prop_name not in query.annotation_select


class TestLegacyOrderingModelIterable(object):

    @pytest.mark.parametrize('select', [
        (),
        ('version_count',),
        ('version_count', 'major_sum'),
    ])
    def test_postprocess_queryable_properties(self, refs, select):
        iterable = LegacyOrderingModelIterable(ApplicationWithClassBasedProperties.objects.all())
        iterable.__dict__['_order_by_select'] = {refs[prop_name] for prop_name in select}
        obj = ApplicationWithClassBasedProperties()
        setattr(obj, QUERYING_PROPERTIES_MARKER, True)
        for ref in iterable._order_by_select:
            ref.descriptor.set_cached_value(obj, 1337)
        obj = iterable._postprocess_queryable_properties(obj)
        for ref in iterable._order_by_select:
            assert not ref.descriptor.has_cached_value(obj)


class TestLegacyValuesIterable(object):

    @pytest.mark.parametrize('prop_names', [
        (),
        ('version_count',),
        ('version_count', 'major_sum'),
    ])
    def test_postprocess_queryable_properties(self, refs, prop_names):
        iterable = LegacyValuesIterable(ApplicationWithClassBasedProperties.objects.all())
        iterable.__dict__['_order_by_select'] = {refs[prop_name] for prop_name in prop_names}
        obj = {'name': 'My cool App', 'version_count': 4, 'major_sum': 5}
        result = iterable._postprocess_queryable_properties(dict(obj))
        assert result == {name: value for name, value in obj.items() if name not in prop_names}


@pytest.mark.skipif(DJANGO_VERSION >= (1, 9), reason='ValuesListQuerySets only exist in old Django versions.')
class TestLegacyValuesListIterable(object):

    @pytest.mark.parametrize('flat', [True, False])
    def test_initializer(self, flat):
        queryset = ApplicationWithClassBasedProperties.objects.values_list('name', flat=flat)
        iterable = LegacyValuesListIterable(queryset)
        assert iterable.queryset.flat is False
        assert iterable.flat is flat

    @pytest.mark.parametrize('select, values, expected_indexes', [
        ((), (), {-1, -2}),
        ((), ('name', 'common_data'), {-1, -2}),
        (('version_count',), (), 'major_sum'),
        (('version_count',), ('version_count', 'name'), {-1}),
        (('major_sum', 'version_count'), (), set()),
        (('major_sum', 'version_count'), ('name',), {-1, -2}),
        (('major_sum', 'version_count'), ('version_count', 'name'), {-1}),
    ])
    def test_discarded_indexes(self, select, values, expected_indexes):
        queryset = ApplicationWithClassBasedProperties.objects.select_properties(*select)
        iterable = LegacyValuesListIterable(queryset.order_by('major_sum', 'version_count').values_list(*values))
        iterable._setup_queryable_properties()
        if not isinstance(expected_indexes, set):
            expected_indexes = {list(iterable.queryset.query.aggregate_select).index(expected_indexes) - 2}
        assert iterable._discarded_indexes == expected_indexes

    @pytest.mark.parametrize('discarded_indexes, expected_result', [
        (set(), tuple(range(10))),
        ({-1}, tuple(range(9))),
        ({-2}, (0, 1, 2, 3, 4, 5, 6, 7, 9)),
        ({-1, -2, -5}, (0, 1, 2, 3, 4, 6, 7)),
    ])
    def test_postprocess_queryable_properties(self, discarded_indexes, expected_result):
        queryset = ApplicationWithClassBasedProperties.objects.values_list('name')
        iterable = LegacyValuesListIterable(queryset)
        iterable.__dict__['_discarded_indexes'] = discarded_indexes
        obj = tuple(range(10))
        assert iterable._postprocess_queryable_properties(obj) == expected_result



# -*- coding: utf-8 -*-

import pytest
import six
from django import VERSION as DJANGO_VERSION
from django.contrib.admin import ModelAdmin, site
from django.contrib.admin.filters import BooleanFieldListFilter, ChoicesFieldListFilter, DateFieldListFilter
from django.contrib.admin.views.main import ChangeList
from django.db.models import DateField
from django.http import QueryDict

from queryable_properties.admin.filters import QueryablePropertyField, QueryablePropertyListFilter
from queryable_properties.exceptions import QueryablePropertyError
from queryable_properties.utils import get_queryable_property
from queryable_properties.utils.internal import QueryPath, get_output_field
from ..app_management.admin import ApplicationAdmin, VersionAdmin
from ..app_management.models import (
    ApplicationWithClassBasedProperties, CategoryWithClassBasedProperties, VersionWithClassBasedProperties,
)


class TestQueryablePropertyField(object):

    @pytest.fixture
    def admin_instance(self):
        return ApplicationAdmin(ApplicationWithClassBasedProperties, site)

    def get_changelist(self, request, model_admin, **kwargs):
        """
        Build a changelist instance for the given admin for testing purposes.

        :param django.http.HttpRequest request: An HTTP request.
        :param model_admin: The model admin to create the changelist for.
        :param kwargs: Further keyword arguments for the changelist
                       constructor, which will otherwise be filled with
                       proper default values.
        :return: The changelist instance.
        :rtype: ChangeList
        """
        list_display = kwargs.get('list_display', model_admin.get_list_display(request))
        defaults = dict(
            model=model_admin.model,
            list_display=list_display,
            list_display_links=model_admin.get_list_display_links(request, list_display),
            list_filter=model_admin.get_list_filter(request) if DJANGO_VERSION >= (1, 5) else model_admin.list_filter,
            date_hierarchy=model_admin.date_hierarchy,
            search_fields=model_admin.search_fields,
            list_select_related=model_admin.list_select_related,
            list_per_page=model_admin.list_per_page,
            list_max_show_all=model_admin.list_max_show_all,
            list_editable=model_admin.list_editable,
            model_admin=model_admin
        )
        if hasattr(ModelAdmin, 'get_sortable_by'):
            defaults['sortable_by'] = model_admin.get_sortable_by(request)
        if hasattr(ModelAdmin, 'search_help_text'):
            defaults['search_help_text'] = model_admin.search_help_text
        defaults.update(kwargs)
        return ChangeList(request, **defaults)

    def get_list_filter_queryset(self, request, field, admin_instance, params):
        if DJANGO_VERSION >= (5, 0):
            query_params = QueryDict(mutable=True)
            query_params.update(params)
        else:
            query_params = params
        list_filter = field.get_filter_creator()(request, query_params, admin_instance.model, admin_instance)
        return list_filter.queryset(request, admin_instance.get_queryset(request))

    @pytest.mark.parametrize('query_path, expected_property', [
        (QueryPath('version_count'), get_queryable_property(ApplicationWithClassBasedProperties, 'version_count')),
        (QueryPath('categories__version_count'),
         get_queryable_property(CategoryWithClassBasedProperties, 'version_count')),
    ])
    def test_initializer(self, admin_instance, query_path, expected_property):
        field = QueryablePropertyField(admin_instance, query_path)
        output_field = get_output_field(expected_property.get_annotation(expected_property.model))
        assert type(field.output_field) is type(output_field)  # Old django versions don't implement field comparison
        assert field.model_admin is admin_instance
        assert field.property is expected_property
        assert field.property_path == query_path
        assert field.null is getattr(output_field, 'null', True)
        assert field.empty_strings_allowed is getattr(output_field, 'null', True)

    @pytest.mark.parametrize('query_path', [
        QueryPath('name'),
        QueryPath('neither_field_nor_property'),
        QueryPath('version_count__gt'),
    ])
    def test_initializer_error(self, admin_instance, query_path):
        with pytest.raises(QueryablePropertyError):
            QueryablePropertyField(admin_instance, query_path)

    def test_attribute_passthrough(self, admin_instance):
        field = QueryablePropertyField(admin_instance, QueryPath('version_count'))
        assert field.model is ApplicationWithClassBasedProperties
        assert field.name == 'version_count'
        assert field.verbose_name == 'Version count'

    @pytest.mark.parametrize('filter_class, expected_filter_class', [
        (None, DateFieldListFilter),
        (ChoicesFieldListFilter, ChoicesFieldListFilter),
    ])
    def test_get_filter_creator(self, rf, admin_instance, filter_class, expected_filter_class):
        field = QueryablePropertyField(admin_instance, QueryPath('support_start_date'))
        field.output_field = DateField(null=True)  # To set an output field for Django versions that don't support it
        creator = field.get_filter_creator(filter_class)
        assert callable(creator)
        list_filter = creator(rf.get('/'), {}, admin_instance.model, admin_instance)
        assert isinstance(list_filter, expected_filter_class)
        assert list_filter.field is field
        assert list_filter.field_path == six.text_type(field.property_path)

    @pytest.mark.django_db
    def test_date_list_filter(self, rf, admin_user, admin_instance):
        field = QueryablePropertyField(admin_instance, QueryPath('support_start_date'))
        field.output_field = DateField(null=True)  # To set an output field for Django versions that don't support it
        request = rf.get('/')
        request.user = admin_user
        list_filter = field.get_filter_creator()(request, {}, admin_instance.model, admin_instance)
        changelist = self.get_changelist(request, admin_instance)
        display_values = [item['display'] for item in list_filter.choices(changelist)]
        expected_values = ['Any date', 'Today', 'Past 7 days', 'This month', 'This year']
        if DJANGO_VERSION >= (1, 10):
            expected_values += ['No date', 'Has date']
        assert display_values == expected_values

    @pytest.mark.django_db
    @pytest.mark.parametrize('params, expected_count', [
        ({}, 2),
        ({'support_start_date__gte': '2016-12-15', 'support_start_date__lt': '2017-01-15'}, 2),
        ({'support_start_date__gte': '2016-12-31', 'support_start_date__lt': '2017-01-31'}, 2),
        ({'support_start_date__gte': '2016-11-15', 'support_start_date__lt': '2016-12-31'}, 0),
        ({'support_start_date__gte': '2018-01-01', 'support_start_date__lt': '2018-12-31'}, 0),
    ])
    @pytest.mark.usefixtures('versions')
    def test_date_list_filter_application(self, rf, admin_user, admin_instance, params, expected_count):
        field = QueryablePropertyField(admin_instance, QueryPath('support_start_date'))
        field.output_field = DateField(null=True)  # To set an output field for Django versions that don't support it
        request = rf.get('/')
        request.user = admin_user
        assert self.get_list_filter_queryset(request, field, admin_instance, params).count() == expected_count

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.django_db
    def test_boolean_list_filter(self, rf, admin_user, admin_instance):
        field = QueryablePropertyField(admin_instance, QueryPath('has_version_with_changelog'))
        assert tuple(field.flatchoices) == ()

        request = rf.get('/')
        request.user = admin_user
        list_filter = field.get_filter_creator()(request, {}, admin_instance.model, admin_instance)
        changelist = self.get_changelist(request, admin_instance)
        display_values = [item['display'] for item in list_filter.choices(changelist)]
        assert display_values == ['All', 'Yes', 'No']

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.django_db
    @pytest.mark.parametrize('params, expected_count', [
        ({}, 2),
        ({'has_version_with_changelog__exact': '0'}, 1),
        ({'has_version_with_changelog__exact': '1'}, 1),
    ])
    def test_boolean_list_filter_application(self, rf, admin_user, versions, admin_instance, params, expected_count):
        versions[3].delete()
        field = QueryablePropertyField(admin_instance, QueryPath('has_version_with_changelog'))
        request = rf.get('/')
        request.user = admin_user
        assert self.get_list_filter_queryset(request, field, admin_instance, params).count() == expected_count

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.django_db
    def test_mapping_choices_list_filter(self, rf, admin_user):
        admin = VersionAdmin(VersionWithClassBasedProperties, site)
        field = QueryablePropertyField(admin, QueryPath('release_type_verbose_name'))
        assert tuple(field.flatchoices) == (
            ('Alpha', 'Alpha'),
            ('Beta', 'Beta'),
            ('Stable', 'Stable'),
            (None, field.empty_value_display),
        )

        request = rf.get('/')
        request.user = admin_user
        list_filter = field.get_filter_creator()(request, {}, admin.model, admin)
        changelist = self.get_changelist(request, admin)
        display_values = [item['display'] for item in list_filter.choices(changelist)]
        assert display_values == ['All', 'Alpha', 'Beta', 'Stable', field.empty_value_display]

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.django_db
    @pytest.mark.parametrize('params, expected_count', [
        ({}, 8),
        ({'release_type_verbose_name__exact': 'Alpha'}, 2),
        ({'release_type_verbose_name__exact': 'Beta'}, 2),
        ({'release_type_verbose_name__exact': 'Stable'}, 4),
    ])
    @pytest.mark.usefixtures('versions')
    def test_mapping_choices_list_filter_application(self, rf, admin_user, params, expected_count):
        admin = VersionAdmin(VersionWithClassBasedProperties, site)
        field = QueryablePropertyField(admin, QueryPath('release_type_verbose_name'))
        request = rf.get('/')
        request.user = admin_user
        assert self.get_list_filter_queryset(request, field, admin, params).count() == expected_count

    @pytest.mark.django_db
    @pytest.mark.parametrize('query_path, expected_choices', [
        (QueryPath('version_count'), ((3, '3'), (4, '4'))),
        (QueryPath('categories__version_count'), ((4, '4'), (7, '7'))),
        pytest.param(QueryPath('major_avg'), ((1.0, '1.0'), (1.25, '1.25')),
                     marks=pytest.mark.skipif(DJANGO_VERSION < (1, 10),
                                              reason="The Cast() expression didn't exist before Django 1.10")),
    ])
    def test_choices_list_filter(self, rf, admin_user, versions, admin_instance, query_path, expected_choices):
        versions[3].delete()
        field = QueryablePropertyField(admin_instance, query_path)
        assert tuple(field.flatchoices) == expected_choices

        request = rf.get('/')
        request.user = admin_user
        list_filter = field.get_filter_creator()(request, {}, admin_instance.model, admin_instance)
        changelist = self.get_changelist(request, admin_instance)
        display_values = [item['display'] for item in list_filter.choices(changelist)]
        assert display_values == ['All'] + [display_value for value, display_value in expected_choices]

    @pytest.mark.django_db
    @pytest.mark.parametrize('params, expected_count', [
        ({}, 2),
        ({'version_count__exact': '3'}, 1),
        ({'version_count__exact': '4'}, 1),
        ({'version_count__exact': '5'}, 0),
    ])
    def test_choices_list_filter_application(self, rf, admin_user, versions, admin_instance, params, expected_count):
        versions[0].delete()
        field = QueryablePropertyField(admin_instance, QueryPath('version_count'))
        request = rf.get('/')
        request.user = admin_user
        assert self.get_list_filter_queryset(request, field, admin_instance, params).count() == expected_count


class TestQueryablePropertyListFilter(object):

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Output fields couldn't be declared before Django 1.8")
    @pytest.mark.parametrize('prop, admin_class, expected_filter_class', [
        (get_queryable_property(ApplicationWithClassBasedProperties, 'has_version_with_changelog'),
         ApplicationAdmin, BooleanFieldListFilter),
        (get_queryable_property(VersionWithClassBasedProperties, 'release_type_verbose_name'),
         VersionAdmin, ChoicesFieldListFilter),
        (get_queryable_property(ApplicationWithClassBasedProperties, 'support_start_date'),
         VersionAdmin, DateFieldListFilter),
        (get_queryable_property(VersionWithClassBasedProperties, 'version'),
         VersionAdmin, ChoicesFieldListFilter),
    ])
    def test_get_class(self, prop, admin_class, expected_filter_class):
        field = QueryablePropertyField(admin_class(prop.model, site), QueryPath(prop.name))
        assert QueryablePropertyListFilter.get_class(field) is expected_filter_class



# encoding: utf-8

from ..utils import MISSING_OBJECT
from .annotation import AggregateProperty, AnnotationProperty, RelatedExistenceCheckProperty
from .base import QueryableProperty, queryable_property
from .cache_behavior import CACHE_RETURN_VALUE, CACHE_VALUE, CLEAR_CACHE, DO_NOTHING
from .mixins import (
    REMAINING_LOOKUPS, AnnotationGetterMixin, AnnotationMixin, LookupFilterMixin, SetterMixin, UpdateMixin,
    boolean_filter, lookup_filter,
)
from .specialized import MappingProperty, RangeCheckProperty, ValueCheckProperty
from .subquery import SubqueryExistenceCheckProperty, SubqueryFieldProperty

__all__ = (
    'MISSING_OBJECT',
    'AggregateProperty', 'AnnotationProperty', 'RelatedExistenceCheckProperty',
    'QueryableProperty', 'queryable_property',
    'CACHE_RETURN_VALUE', 'CACHE_VALUE', 'CLEAR_CACHE', 'DO_NOTHING',
    'AnnotationGetterMixin', 'AnnotationMixin', 'boolean_filter', 'LookupFilterMixin', 'lookup_filter',
    'REMAINING_LOOKUPS', 'SetterMixin', 'UpdateMixin',
    'MappingProperty', 'RangeCheckProperty', 'ValueCheckProperty',
    'SubqueryExistenceCheckProperty', 'SubqueryFieldProperty',
)



# encoding: utf-8

import pytest
from django import VERSION as DJANGO_VERSION
from django.db import connection, models

from queryable_properties.query import QUERYING_PROPERTIES_MARKER
from queryable_properties.utils.internal import get_queryable_property_descriptor
from ..app_management.models import (
    ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties, VersionWithClassBasedProperties,
    VersionWithDecoratorBasedProperties,
)
from ..conftest import Concat, Value

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestAggregateAnnotations(object):

    @pytest.mark.parametrize('model, filters', [
        (ApplicationWithClassBasedProperties, {}),
        (ApplicationWithClassBasedProperties, {'version_count__gt': 3}),
        (ApplicationWithDecoratorBasedProperties, {}),
        (ApplicationWithDecoratorBasedProperties, {'version_count__gt': 3}),
    ])
    def test_cached_annotation_value(self, model, filters):
        # Filter both before and after the select_properties call to check if
        # the annotation gets selected correctly regardless
        queryset = model.objects.filter(**filters).select_properties('version_count', 'major_sum').filter(**filters)
        assert 'version_count' in queryset.query.annotations
        assert 'major_sum' in queryset.query.annotations
        for application in queryset:
            assert model.version_count.has_cached_value(application)
            assert application.version_count == 4
            assert model.major_sum.has_cached_value(application)
            assert application.major_sum == 5
            assert not hasattr(application, QUERYING_PROPERTIES_MARKER)

    @pytest.mark.parametrize('model, limit, expected_total', [
        (ApplicationWithClassBasedProperties, None, 8),
        (ApplicationWithClassBasedProperties, 1, 4),
        (ApplicationWithDecoratorBasedProperties, None, 8),
        (ApplicationWithDecoratorBasedProperties, 1, 4),
    ])
    def test_aggregate_based_on_queryable_property(self, model, limit, expected_total):
        result = model.objects.all()[:limit].aggregate(total_version_count=models.Sum('version_count'))
        assert result['total_version_count'] == expected_total

    @pytest.mark.parametrize('model, limit, expected_total', [
        (VersionWithClassBasedProperties, None, 32),
        (VersionWithClassBasedProperties, 4, 16),
        (VersionWithDecoratorBasedProperties, None, 32),
        (VersionWithDecoratorBasedProperties, 4, 16),
    ])
    def test_aggregate_based_on_queryable_property_across_relation(self, model, limit, expected_total):
        result = model.objects.all()[:limit].aggregate(total_version_count=models.Sum('application__version_count'))
        assert result['total_version_count'] == expected_total

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.parametrize('model, annotation', [
        (VersionWithClassBasedProperties, models.F('application__version_count')),
        (VersionWithDecoratorBasedProperties, models.F('application__version_count')),
    ])
    def test_annotation_based_on_queryable_property_across_relation(self, model, annotation):
        model.objects.all()[0].delete()
        queryset = model.objects.annotate(annotation=annotation)
        assert all(obj.annotation == obj.application.version_count for obj in queryset)

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_iterator(self, model):
        queryset = model.objects.filter(version_count=4).select_properties('version_count')
        for application in queryset.iterator():
            assert model.version_count.has_cached_value(application)
            assert application.version_count == 4
        assert queryset._result_cache is None

    @pytest.mark.skipif(DJANGO_VERSION < (3, 2), reason="The alias() method didn't exist before Django 3.2")
    @pytest.mark.parametrize('model, with_selection', [
        (ApplicationWithClassBasedProperties, False),
        (ApplicationWithClassBasedProperties, True),
        (ApplicationWithDecoratorBasedProperties, False),
        (ApplicationWithDecoratorBasedProperties, True),
    ])
    def test_alias(self, model, with_selection):
        queryset = model.objects.alias(alias=models.F('version_count') + 1)
        if with_selection:
            queryset = queryset.select_properties('version_count')
        results = list(queryset.filter(alias=5))
        assert len(results) == 2
        for application in results:
            assert model.version_count.has_cached_value(application) is with_selection
            assert application.version_count == 4

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_removed_annotation(self, model):
        """
        Test that queries can still be performed even if queryable property
        annotations have been manually removed from the queryset.
        """
        queryset = model.objects.select_properties('version_count')
        del queryset.query.annotations['version_count']
        assert bool(queryset)
        assert all(not model.version_count.has_cached_value(obj) for obj in queryset)

    @pytest.mark.skipif(DJANGO_VERSION < (1, 7), reason="Raw queries didn't exist before Django 1.7")
    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_raw_query(self, model):
        pks, names = zip(*model.objects.values_list('pk', 'name'))
        queryset = model.objects.raw('SELECT id, name, 5 AS version_count FROM {}'.format(model._meta.db_table))
        counter = 0
        for application in queryset:
            assert model.version_count.has_cached_value(application) is True
            assert application.version_count == 5
            assert application.pk in pks
            assert application.name in names
            assert not hasattr(application, QUERYING_PROPERTIES_MARKER)
            counter += 1
        assert counter == 2


@pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
class TestExpressionAnnotations(object):

    @pytest.mark.parametrize('model, filters', [
        (VersionWithClassBasedProperties, {}),
        (VersionWithDecoratorBasedProperties, {}),
        (VersionWithClassBasedProperties, {'version': '1.2.3'}),
        (VersionWithDecoratorBasedProperties, {'version': '1.2.3'}),
    ])
    def test_cached_annotation_value(self, model, filters):
        # Filter both before and after the select_properties call to check if
        # the annotation gets selected correctly regardless
        queryset = model.objects.filter(**filters).select_properties('version').filter(**filters)
        assert 'version' in queryset.query.annotations
        assert all(model.version.has_cached_value(obj) for obj in queryset)

    @pytest.mark.parametrize('model, property_name, annotation, expected_count, record_checker', [
        (VersionWithClassBasedProperties, 'version', models.F('version'), 8,
         lambda obj: obj.annotation == obj.version),
        (VersionWithDecoratorBasedProperties, 'version', models.F('version'), 8,
         lambda obj: obj.annotation == obj.version),
        (VersionWithClassBasedProperties, 'version', Concat(Value('V'), 'version'), 8,
         lambda obj: obj.annotation == 'V' + obj.version),
        (VersionWithDecoratorBasedProperties, 'version', Concat(Value('V'), 'version'), 8,
         lambda obj: obj.annotation == 'V' + obj.version),
        (ApplicationWithClassBasedProperties, 'versions__version', models.F('versions__version'), 8,
         lambda obj: obj.annotation in ('1.2.3', '1.3.0', '1.3.1', '2.0.0')),
        (ApplicationWithDecoratorBasedProperties, 'versions__version', models.F('versions__version'), 8,
         lambda obj: obj.annotation in ('1.2.3', '1.3.0', '1.3.1', '2.0.0')),
        (ApplicationWithClassBasedProperties, 'versions__version', Concat(Value('V'), 'versions__version'), 8,
         lambda obj: obj.annotation in ('V1.2.3', 'V1.3.0', 'V1.3.1', 'V2.0.0')),
        (ApplicationWithDecoratorBasedProperties, 'versions__version', Concat(Value('V'), 'versions__version'), 8,
         lambda obj: obj.annotation in ('V1.2.3', 'V1.3.0', 'V1.3.1', 'V2.0.0')),
    ])
    def test_annotation_based_on_queryable_property(self, model, property_name, annotation, expected_count,
                                                    record_checker):
        queryset = model.objects.annotate(annotation=annotation)
        assert queryset.count() == len(queryset) == expected_count
        assert all(record_checker(obj) for obj in queryset)
        if '__' not in property_name:
            # Check that a property annotation used implicitly by another
            # annotation does not lead to a selection of the property
            # annotation
            descriptor = get_queryable_property_descriptor(model, property_name)
            assert all(not descriptor.has_cached_value(obj) for obj in queryset)

    @pytest.mark.skipif(DJANGO_VERSION < (3, 2), reason="The alias() method didn't exist before Django 3.2")
    @pytest.mark.parametrize('model, with_selection', [
        (VersionWithClassBasedProperties, False),
        (VersionWithClassBasedProperties, True),
        (VersionWithDecoratorBasedProperties, False),
        (VersionWithDecoratorBasedProperties, True),
    ])
    def test_alias(self, model, with_selection):
        queryset = model.objects.alias(alias=Concat(Value('V'), 'version'))
        if with_selection:
            queryset = queryset.select_properties('version')
        results = list(queryset.filter(alias='V2.0.0'))
        assert len(results) == 2
        for version in results:
            assert model.version.has_cached_value(version) is with_selection
            assert version.version == '2.0.0'

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_iterator(self, model):
        queryset = model.objects.filter(major_minor='2.0').select_properties('version')
        for version in queryset.iterator():
            assert model.version.has_cached_value(version)
            assert version.version == '2.0.0'
        assert queryset._result_cache is None

    @pytest.mark.skipif(DJANGO_VERSION < (4, 2), reason='Sliced prefetches were introduced in Django 4.2')
    def test_sliced_prefetch(self):
        from django.test.utils import CaptureQueriesContext

        with CaptureQueriesContext(connection) as context:
            versions = VersionWithClassBasedProperties.objects.select_properties('released_in_2018').order_by(
                'major', 'minor', 'patch')
            apps = ApplicationWithClassBasedProperties.objects.prefetch_related(
                models.Prefetch('versions', versions[1:3], to_attr='prefetched_versions'))
            for app in apps:
                assert all(version.major_minor == '1.3' for version in app.prefetched_versions)
            assert context.captured_queries[1]['sql'].count('"released_in_2018"') == 2



# -*- coding: utf-8 -*-

from collections import OrderedDict

import six
from django.contrib.admin.filters import (
    BooleanFieldListFilter, ChoicesFieldListFilter, DateFieldListFilter, FieldListFilter,
)
from django.contrib.admin.views import main
from django.db.models import BooleanField, DateField

from ..exceptions import QueryablePropertyError
from ..managers import QueryablePropertiesQuerySet
from ..properties import MappingProperty
from ..utils.internal import QueryPath, get_output_field, resolve_queryable_property


class QueryablePropertyField(object):
    """
    Wrapper class for queryable property that offers an attribute interface
    similar to Django fields. This allows to reuse Django's existing list
    filter implementations for queryable properties by providing objects of
    this class to list filters.
    """

    def __init__(self, model_admin, query_path):
        """
        Initialize a new property-to-field interface wrapper for the queryable
        property with the given path.

        :param model_admin: The admin instance for which a filter based on the
                            given queryable property should be created.
        :type model_admin: django.contrib.admin.options.BaseModelAdmin
        :param QueryPath query_path: The query path to the queryable property.
        """
        property_ref, lookups = resolve_queryable_property(model_admin.model, query_path)
        if not property_ref or lookups:
            raise QueryablePropertyError('The query path must point to a valid queryable property and may not contain'
                                         'lookups/transforms.')

        self.output_field = get_output_field(property_ref.get_annotation())
        self.model_admin = model_admin
        self.property = property_ref.property
        self.property_ref = property_ref
        self.property_path = query_path
        self.null = self.output_field is None or self.output_field.null
        self.empty_strings_allowed = self.output_field is None or self.output_field.empty_strings_allowed

    def __getattr__(self, item):
        # Pass through attribute accesses to the associated queryable property.
        # This allows to access properties common to both queryable properties
        # and model fields (e.g. name, verbose_name) without having to copy
        # their values explicitly.
        return getattr(self.property, item)

    @property
    def empty_value_display(self):
        """
        Get the value to display in the admin in place of empty values.

        :return: The value to display for empty internal values.
        :rtype: str
        """
        return getattr(main, 'EMPTY_CHANGELIST_VALUE', None) or self.model_admin.get_empty_value_display()

    @property
    def flatchoices(self):
        """
        Build the list filter choices for the associated queryable property
        as 2-tuples.

        This is an attribute expected by certain list filter classes and is
        used for any queryable property that doesn't map to a specialized
        filter class, which is why the :class:`ChoicesFieldListFilter` is used
        as the last resort when determining filter classes.

        :return: The filter choices as 2-tuples containing the internal value
                 as the first and the display value as the second item.
        :rtype: (object, object)
        """
        if isinstance(self.property, MappingProperty):
            options = OrderedDict((to_value, to_value) for from_value, to_value in self.property.mappings)
            options.setdefault(self.property.default, self.empty_value_display)
            for value, label in six.iteritems(options):
                yield value, label
        elif not isinstance(self.output_field, BooleanField):
            name = six.text_type(QueryPath(('', 'value')))
            queryset = QueryablePropertiesQuerySet.get_for_model(self.property_ref.model)
            queryset = queryset.annotate(**{name: self.property_ref.get_annotation()}).order_by(name).distinct()
            for value in queryset.values_list(name, flat=True):
                yield value, six.text_type(value) if value is not None else self.empty_value_display

    def get_filter_creator(self, list_filter_class=None):
        """
        Create a callable that can be used to create a list filter object based
        on this property.

        :param list_filter_class: The list filter class to use. If not given,
                                  a suitable list filter class will be
                                  determined for the associated queryable
                                  property.
        :return: A callable to create a list filter object.
        :rtype: collections.Callable
        """
        list_filter_class = list_filter_class or QueryablePropertyListFilter.get_class(self)

        def creator(request, params, model, model_admin):
            return list_filter_class(self, request, params, model, model_admin, six.text_type(self.property_path))
        return creator


class QueryablePropertyListFilter(FieldListFilter):
    """
    A base list filter class for queryable properties that allows to re-use
    Django's filter class registration for queryable properties.
    """
    _field_list_filters = []
    _take_priority_index = 0

    @classmethod
    def get_class(cls, field):
        """
        Determine a suitable list filter class for the given wrapped queryable
        property based on the registered filter classes.

        :param QueryablePropertyField field: The wrapped queryable property.
        :return: An appropriate list filter class.
        """
        for test, list_filter_class in cls._field_list_filters:
            if test(field):
                return list_filter_class


QueryablePropertyListFilter.register(lambda field: isinstance(field.output_field, BooleanField), BooleanFieldListFilter)
QueryablePropertyListFilter.register(lambda field: isinstance(field.property, MappingProperty), ChoicesFieldListFilter)
QueryablePropertyListFilter.register(lambda field: isinstance(field.output_field, DateField), DateFieldListFilter)
# Use the ChoicesFieldListFilter as the last resort since the general
# implementation of "list all possible values" is implemented in
# `QueryablePropertyField.flatchoices`.
# Django's last resort AllValuesFieldListFilter cannot be used as it performs
# a hardcoded query on its own, which wouldn't work with a queryable property.
QueryablePropertyListFilter.register(lambda field: True, ChoicesFieldListFilter)



# encoding: utf-8

from __future__ import unicode_literals

from copy import deepcopy
from functools import partial

import six

from ..compat import LOOKUP_SEP, pretty_name
from ..exceptions import QueryablePropertyError
from ..query import QUERYING_PROPERTIES_MARKER
from ..utils import get_queryable_property, reset_queryable_property
from ..utils.internal import parametrizable_decorator_method
from .cache_behavior import CLEAR_CACHE
from .mixins import AnnotationGetterMixin, AnnotationMixin, LookupFilterMixin

RESET_METHOD_NAME = 'reset_property'


@six.python_2_unicode_compatible
class QueryablePropertyDescriptor(property):
    """
    Descriptor class for queryable properties that allows the actual attribute
    access on model instances and handles caching.

    This class deliberately inherits from property to be treated like regular
    properties by Django, e.g. to allow queryable properties with a setter to
    be used in the initializer kwargs of models.
    """

    def __new__(cls, prop):
        """
        Construct a new QueryablePropertyDescriptor for the given queryable
        property.

        :param prop: The queryable property to allow attribute access for.
        :type prop: QueryableProperty
        """
        descriptor = super(QueryablePropertyDescriptor, cls).__new__(cls, doc=prop.__doc__)
        descriptor.prop = prop
        return descriptor

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        # Always check for cached values first regardless of the associated
        # property being configured as cached since values will also be cached
        # through annotation selections.
        if self.has_cached_value(obj):
            return self.get_cached_value(obj)
        if not self.prop.get_value:
            raise AttributeError('Unreadable queryable property.')
        value = self.prop.get_value(obj)
        if self.prop.cached:
            self.set_cached_value(obj, value)
        return value

    def __set__(self, obj, value):
        # Values set while initializing new objects from DB values should
        # always be cached regardless of the actual setter.
        if getattr(obj, QUERYING_PROPERTIES_MARKER, False):
            self.set_cached_value(obj, value)
            return
        if not self.prop.set_value:
            raise AttributeError("Can't set queryable property.")
        return_value = self.prop.set_value(obj, value)
        # If a value is set and the property is set up to cache values or has
        # a current cached value, invoke the configured setter cache behavior.
        if self.prop.cached or self.has_cached_value(obj):
            self.prop.setter_cache_behavior(self, obj, value, return_value)

    def __str__(self):
        return six.text_type(self.prop)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, six.text_type(self))

    def get_cached_value(self, obj):
        """
        Get the cached value for the associated queryable property from the
        given object. Requires a cached value to be present.

        :param django.db.models.Model obj: The object to get the cached value
                                           from.
        :return: The cached value.
        """
        return obj.__dict__[self.prop.name]

    def set_cached_value(self, obj, value):
        """
        Set the cached value for the associated queryable property on the given
        object.

        :param django.db.models.Model obj: The object to set the cached value
                                           for.
        :param value: The value to cache.
        """
        obj.__dict__[self.prop.name] = value

    def has_cached_value(self, obj):
        """
        Check if a value for the associated queryable property is cached on the
        given object.

        :param django.db.models.Model obj: The object to check for a cached
                                           value.
        :return: True if a value is cached; otherwise False.
        :rtype: bool
        """
        return self.prop.name in obj.__dict__

    def clear_cached_value(self, obj):
        """
        Clear the cached value for the associated queryable property on the
        given object. Does not require a cached value to be present and will
        do nothing if no value is cached.

        :param django.db.models.Model obj: The object to clear the cached value
                                           on.
        """
        obj.__dict__.pop(self.prop.name, None)


@six.python_2_unicode_compatible
class QueryableProperty(object):
    """
    Base class for all queryable property definitions, which provide methods
    for single object as well as queryset interaction.
    """

    cached = False  #: Determines if the result of the getter is cached, like Python's/Django's ``cached_property``.
    setter_cache_behavior = CLEAR_CACHE  #: Determines what happens if the setter of a cached property is used.
    filter_requires_annotation = False  #: Determines if using the property to filter requires annotating first.

    # Set the attributes of mixin methods to None for easier checks if a
    # property implements them.
    set_value = None
    get_annotation = None
    get_update_kwargs = None

    def __init__(self, verbose_name=None):
        """
        Initialize a new queryable property.

        :param str verbose_name: An optional verbose name of the property. If
                                 not provided it defaults to a prettified
                                 version of the property's name.
        """
        self.model = None
        self.name = None
        self.setter_cache_behavior = six.get_method_function(self.setter_cache_behavior)
        self.verbose_name = verbose_name

    def __reduce__(self):
        # Since queryable property instances only make sense in the context of
        # model classes, they can simply be pickled using their model class and
        # name and loaded back from the model class when unpickling. This also
        # saves memory as unpickled properties will be the exact same object as
        # the one on the model class.
        return get_queryable_property, (self.model, self.name)

    def __str__(self):
        return '.'.join((self.model.__module__, self.model.__name__, self.name))

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, six.text_type(self))

    @property
    def short_description(self):
        """
        Return the verbose name of this property as its short description,
        which is required for the admin integration.

        :return: The verbose name of this property.
        :rtype: str
        """
        return self.verbose_name

    def get_value(self, obj):  # pragma: no cover
        """
        Getter method for the queryable property, which will be called when
        the property is read-accessed.

        :param django.db.models.Model obj: The object on which the property was accessed.
        :return: The getter value.
        """
        raise NotImplementedError()

    def get_filter(self, cls, lookup, value):  # pragma: no cover
        """
        Generate a :class:`django.db.models.Q` object that emulates filtering
        a queryset using this property.

        :param type cls: The model class of which a queryset should be
                         filtered.
        :param str lookup: The lookup to use for the filter (e.g. 'exact',
                           'lt', etc.)
        :param value: The value passed to the filter condition.
        :return: A Q object to filter using this property.
        :rtype: django.db.models.Q
        """
        raise NotImplementedError()

    def contribute_to_class(self, cls, name):
        if LOOKUP_SEP in name:
            raise QueryablePropertyError('The name of a queryable property must not contain the lookup separator "{}".'
                                         .format(LOOKUP_SEP))
        # Store some useful values on model class initialization.
        self.model = self.model or cls
        self.name = self.name or name
        if self.verbose_name is None:
            self.verbose_name = pretty_name(self.name)
        setattr(cls, name, QueryablePropertyDescriptor(self))  # Add a descriptor for this property to the model class
        # If not already set, also add a method to the model class that allows
        # to reset the cached values of queryable properties.
        if not getattr(cls, RESET_METHOD_NAME, None):
            setattr(cls, RESET_METHOD_NAME, reset_queryable_property)


class queryable_property(QueryableProperty):
    """
    A queryable property that is intended to be used as a decorator.
    """

    # Set the attributes of the default methods to None since the decorator
    # may be used without implementing these methods.
    get_value = None
    get_filter = None

    def __init__(self, getter=None, cached=None, annotation_based=False, **kwargs):
        """
        Initialize a new queryable property, optionally using the given getter
        method and getter configuration.

        :param function getter: The method to decorate.
        :param cached: Determines if values obtained by the getter should be
                       cached (similar to ``cached_property``). A value of None
                       means using the default value.
        :type cached: bool | None
        :param annotation_based: If True, the :class:`AnnotationGetterMixin` is
                                 automatically added to this property to define
                                 a getter implementation and this property is
                                 expected to decorate the annotater method
                                 instead of the getter. If False, property is
                                 expected to decorate the getter method.
        :type annotation_based: bool
        """
        super(queryable_property, self).__init__(**kwargs)
        self.__doc__ = None
        if getter:
            self(getter, force_getter=True)
        if cached is not None:
            self.cached = cached
        if annotation_based:
            AnnotationGetterMixin.inject_into_object(self)

    def __call__(self, method, force_getter=False):
        # Since the initializer may be used as a parametrized decorator, the
        # resulting object will be called to apply the decorator.
        if force_getter or not isinstance(self, AnnotationGetterMixin):
            self.get_value = method
        else:
            self.get_annotation = self._extract_function(method)
        self.__doc__ = method.__doc__ or self.__doc__
        return self

    def _extract_function(self, method_or_function):
        """
        Extract the function from the given function or method. Allows to
        decorate either regular functions or e.g. classmethods with the
        decorators of this property.

        :param method_or_function: The decorated method or function.
        :type method_or_function: function | classmethod | staticmethod
        :return: The actual function object.
        """
        return getattr(method_or_function, '__func__', method_or_function)

    def _clone(self, **kwargs):
        """
        Clone this queryable property while overriding attributes. This is
        necessary whenever an additional decorator is used to not mess up in
        inheritance scenarios.

        :param kwargs: Attributes to override.
        :return: A (modified) clone of this queryable property.
        :rtype: queryable_property
        """
        attrs = deepcopy(self.__dict__)
        attrs.update(kwargs)
        clone = self.__class__()
        clone.__dict__.update(attrs)
        return clone

    @parametrizable_decorator_method
    def getter(self, method, cached=None):
        """
        Decorator for a function or method that is used as the getter of this
        queryable property. May be used as a parameter-less decorator
        (``@getter``) or as a decorator with keyword arguments
        (``@getter(cached=True)``).

        :param function method: The method to decorate.
        :param cached: If ``True``, values returned by the decorated getter
                       method will be cached. A value of None means no change.
        :type cached: bool | None
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        clone = self._clone()
        if cached is not None:
            clone.cached = cached
        return clone(method, force_getter=True)

    @parametrizable_decorator_method
    def setter(self, method, cache_behavior=None):
        """
        Decorator for a function or method that is used as the setter of this
        queryable property. May be used as a parameter-less decorator
        (``@setter``) or as a decorator with keyword arguments
        (``@setter(cache_behavior=DO_NOTHING)``).

        :param function method: The method to decorate.
        :param cache_behavior: A function that defines how the setter interacts
                               with cached values. A value of None means no
                               change.
        :type cache_behavior: function | None
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        attrs = dict(set_value=method)
        if cache_behavior:
            attrs['setter_cache_behavior'] = cache_behavior
        return self._clone(**attrs)

    @parametrizable_decorator_method
    def filter(self, method, requires_annotation=None, lookups=None, boolean=False, remaining_lookups_via_parent=None):
        """
        Decorator for a function or method that is used to generate a filter
        for querysets to emulate filtering by this queryable property. May be
        used as a parameter-less decorator (``@filter``) or as a decorator with
        keyword arguments (``@filter(requires_annotation=False)``). May be used
        to define a one-for-all filter function or a filter function that will
        be called for certain lookups only using the ``lookups`` argument.

        :param method: The method to decorate.
        :type method: function | classmethod | staticmethod
        :param requires_annotation: ``True`` if filtering using this queryable
                                    property requires its annotation to be
                                    applied first; otherwise ``False``. None if
                                    this information should not be changed.
        :type requires_annotation: bool | None
        :param lookups: If given, the decorated function or method will be used
                        for the specified lookup(s) only. Automatically adds
                        the :class:`LookupFilterMixin` to this property if this
                        is used.
        :type lookups: collections.Iterable[str] | None
        :param boolean: If ``True``, the decorated function or method is
                        expected to be a simple boolean filter, which doesn't
                        take the ``lookup`` and ``value`` parameters and should
                        always return a ``Q`` object representing the positive
                        (i.e. ``True``) filter case. The decorator will
                        automatically negate the condition if the filter was
                        called with a ``False`` value.
        :type boolean: bool
        :param remaining_lookups_via_parent: ``True`` if lookup-based filters
                                             should fall back to the base class
                                             implementation for lookups without
                                             a registered filter function;
                                             otherwise ``False``. None if this
                                             information should not be changed.
        :type remaining_lookups_via_parent: bool
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        method = extracted = self._extract_function(method)
        if boolean:
            if lookups is not None:
                raise QueryablePropertyError('A boolean filter cannot specify lookups at the same time.')

            # Re-use the boolean_filter decorator by simulating a method with
            # a self argument when in reality `method` doesn't have one.
            method = LookupFilterMixin.boolean_filter(lambda prop, model: extracted(model))
            lookups = method._lookups
            method = partial(method, None)

        if remaining_lookups_via_parent is not None and not hasattr(self, 'lookup_mappings') and lookups is None:
            raise QueryablePropertyError('remaining_lookups_via_parent can only be used with lookup-based filters.')

        attrs = {}
        if requires_annotation is not None:
            attrs['filter_requires_annotation'] = requires_annotation
        if remaining_lookups_via_parent is not None:
            attrs['remaining_lookups_via_parent'] = remaining_lookups_via_parent
        if lookups is not None:  # Register only for the given lookups.
            attrs['lookup_mappings'] = dict(getattr(self, 'lookup_mappings', {}),
                                            **{lookup: method for lookup in lookups})
        else:  # Register as a one-for-all filter function.
            attrs['get_filter'] = method
        clone = self._clone(**attrs)
        # If the decorated function/method is used for certain lookups only,
        # add the LookupFilterMixin into the new property to be able to reuse
        # its filter implementation based on the lookup mappings.
        if lookups is not None:
            LookupFilterMixin.inject_into_object(clone)
        return clone

    def annotater(self, method):
        """
        Decorator for a function or method that is used to generate an
        annotation to represent this queryable property in querysets. The
        :class:`AnnotationMixin` will automatically applied to this property
        when this decorator is used.

        :param method: The method to decorate.
        :type method: function | classmethod | staticmethod
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        clone = self._clone(get_annotation=self._extract_function(method))
        # Dynamically add the AnnotationMixin into the new property to allow
        # to use the default filter implementation. Since an explicitly set
        # filter implementation is stored in the instance dict, it will be used
        # over the default implementation.
        return AnnotationMixin.inject_into_object(clone)

    def updater(self, method):
        """
        Decorator for a function or method that is used to resolve an update
        keyword argument for this queryable property into the actual update
        keyword arguments.

        :param method: The method to decorate.
        :type method: function | classmethod | staticmethod
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        return self._clone(get_update_kwargs=self._extract_function(method))
